//
// Created by trokay on 13.06.23.
//

#include <features.h>
#include <linux/if.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <linux/route.h>
/* #include <net/if.h> */
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <ifaddrs.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h> // write(), read(), close()
#include <sys/socket.h>
#include <ctype.h>
// ioctl
#include <stddef.h>
/* #include <net/route.h> */
#include <syslog.h>
#include <netinet/in.h>
#include "network.h"


#define BUFSIZE 8192
#define ifreq_offsetof(x)  offsetof(struct ifreq, x)

const char *DNS_command = "cat /etc/resolv.conf|grep -im 1 '^nameserver' |cut -d ' ' -f2";
const char *gateway_command = "ip route show default | awk '{print $3}'";

const char *RESOLV_FILE = "/etc/resolv.conf";
const char *INTERFACE_FILE = "/etc/network/interfaces";

int parse_ip_address(char *ip_string, uint32_t *numbers);

char check_dhcp_enabled();

int delNullRoute(const char *address);

int execute_command(const char *command, char string[16]);

struct ifaddrs *test;

int init_network_interface(char *interface_name) {
    struct ifaddrs *addresses;
    if (getifaddrs(&addresses) == -1) {
        printf("getifaddrs call failed\n");
        return 1;
    }

    struct ifaddrs *ifa = addresses;
    while (ifa != NULL) {
        if (ifa->ifa_addr != NULL
            && ifa->ifa_addr->sa_family == AF_INET
            && ifa->ifa_netmask != NULL
            && ifa->ifa_flags & IFF_UP
            && strcmp(ifa->ifa_name, interface_name) == 0
                ) {
            test = ifa;
            freeifaddrs(addresses);
            return 0;
        }
        ifa = ifa->ifa_next;
    }
    freeifaddrs(addresses);
    return 1;
}

//int set_c(uint8_t *dhcp, uint32_t *ip, uint32_t *netmask, uint32_t *gw, uint32_t *dns, char *interface_name){
//
//    return 1;
//}

char check_dhcp_enabled() {
    FILE *file = fopen(INTERFACE_FILE, "r");
    if (file == NULL) {
        printf("Failed to open file: %s\n", INTERFACE_FILE);
        return -1;
    }

    char line[256];
    while (fgets(line, sizeof(line), file) != NULL) {
        if (strstr(line, "dhcp") != NULL) {
            fclose(file);
            return 1; // DHCP enabled
        }
    }

    if (fclose(file) != 0) {
        perror("Failed to close file");
        return -2;
    }
    return 0; // DHCP disabled
}

int get_dns(uint32_t *dns) {
    char dd[INET_ADDRSTRLEN];
    uint32_t tmp;
    execute_command(DNS_command, dd);
    parse_ip_address(dd, &tmp);
    memcpy(dns, &tmp, sizeof(uint32_t));
    return 1;
}


int set_dns(uint32_t dns) {
    uint8_t *tmp;
    char msg[50];
    tmp = (uint8_t *)&dns;
    FILE *file = fopen(RESOLV_FILE, "w");
    if (file == NULL) {
        perror("Failed to open file");
        return -1;
    }

    snprintf(msg, 50, "nameserver %d.%d.%d.%d", tmp[0], tmp[1], tmp[2], tmp[3]);
    //const char *msg = "nameserver ";
    fwrite(msg, sizeof(char), strlen(msg), file);
    //fwrite(tmp, sizeof(char), strlen(tmp), file);

    if (fclose(file) != 0) {
        perror("Failed to close file");
        return -2;
    }

    printf("File written and closed successfully.\n");
    return 0;
}

int delNullRoute(const char *address) {
    //int fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_IP);
    //int fd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP);

    printf("Deleting route to %s\n", address);
    int fd = socket(AF_INET, SOCK_DGRAM, 0);

    struct rtentry route;
    memset(&route, 0, sizeof(route));

    struct sockaddr_in *addr = (struct sockaddr_in *) &route.rt_gateway;
    addr->sin_family = AF_INET;
    addr->sin_addr.s_addr = inet_addr(address);

    addr = (struct sockaddr_in *) &route.rt_dst;
    addr->sin_family = AF_INET;
    addr->sin_addr.s_addr = inet_addr("0.0.0.0");  // Replace with the correct destination IP address

    addr = (struct sockaddr_in *) &route.rt_genmask;
    addr->sin_family = AF_INET;
    addr->sin_addr.s_addr = inet_addr("0.0.0.0");

    route.rt_flags = RTF_UP | RTF_GATEWAY;
    //route.rt_metric = 0;

    int result = ioctl(fd, SIOCDELRT, &route);
    if (result < 0) {
        perror("ioctl");
        close(fd);
        return 0;
    }

    close(fd);
    return 1;
}

int execute_command(const char *command, char result[16]) {
    char buffer[128];
    FILE *pipe = popen(command, "r");

    if (pipe == NULL) {
        printf("Failed to execute command\n");
        return 0;
    }

    // Read command output line by line
    if (fgets(buffer, sizeof(buffer), pipe) != NULL) {
        // Remove newline character
        size_t len = strlen(buffer);
        if (len > 0 && buffer[len - 1] == '\n')
            buffer[len - 1] = '\0';

        // Allocate memory for the result string
       // size_t result_len = strlen(buffer);

        // Concatenate the current line to the result string
        for (int i = 0; i < INET_ADDRSTRLEN; ++i) {
            result[i] = buffer[i];
        }
    }

    pclose(pipe);
    return 1;
}

int parse_ip_address(char *ip_string, uint32_t *numbers) {
    char *token = strtok(ip_string, ".");
    uint8_t *p;
    p = (uint8_t *)numbers;
    int i = 0;
    while (token != NULL && i < 4) {
        p[i] = atoi(token);
        token = strtok(NULL, ".");
        i++;
    }
    if (i != 4) {
        printf("Invalid IP address format.\n");
        return -1;
    }
    return 0;
}


static int get_iface_param(const char *ifname, void *data, int op)
{
    int fd = socket(AF_INET, SOCK_DGRAM, 0);
    if(fd == -1) {
	syslog(LOG_ERR, "cannot create socket: %s", strerror(errno));
	return -1;
    }

    struct ifreq ifr;
    ifr.ifr_addr.sa_family = AF_INET;
    strncpy(ifr.ifr_name, ifname, IFNAMSIZ-1);
    if(ioctl(fd, op, &ifr) == -1) {
	close(fd);
	return -1;
    }
    close(fd);

    struct sockaddr_in *addr = (struct sockaddr_in *)&ifr.ifr_addr;
    if(op == SIOCGIFHWADDR)
	memcpy(data, ifr.ifr_hwaddr.sa_data, 6);
    else
	*(uint32_t *)data = addr->sin_addr.s_addr;
    return 0;
}


static int set_iface_param(const char *ifname, uint32_t ip, int op)
{
    int fd = socket(AF_INET, SOCK_DGRAM, 0);
    if(fd == -1) {
	return -1;
    }

    struct ifreq ifr;
    ifr.ifr_addr.sa_family = AF_INET;
    struct sockaddr_in *addr = (struct sockaddr_in *)&ifr.ifr_addr;
    addr->sin_addr.s_addr = ip;
    strncpy(ifr.ifr_name, ifname, IFNAMSIZ-1);
    if(ioctl(fd, op, &ifr) == -1) {
	close(fd);
	return -1;
    }
    close(fd);

    return 0;
}
/**
 * \brief Get interface MAC address
 */
int get_iface_mac(const char *ifname, uint8_t *mac)
{
    return get_iface_param(ifname, mac, SIOCGIFHWADDR);
}

/**
 * \brief Get interface IP address
 */
int get_iface_ip(const char *ifname, uint32_t *ip)
{
    return get_iface_param(ifname, ip, SIOCGIFADDR);
}

int set_iface_ip(const char *ifname, uint32_t ip)
{
    return set_iface_param(ifname, ip, SIOCSIFADDR);
}

int get_iface_mask(const char *ifname, uint32_t *msk)
{
    return get_iface_param(ifname, msk, SIOCGIFNETMASK);
}

int set_iface_mask(const char *ifname, uint32_t msk)
{
    return set_iface_param(ifname, msk, SIOCSIFNETMASK);
}

int get_iface_peer(const char *ifname, uint32_t *ip)
{
    return get_iface_param(ifname, ip, SIOCGIFDSTADDR);
}



int iface_up(const char *ifname)
{
    int fd = socket(AF_INET, SOCK_DGRAM, 0);
    if(fd == -1) {
	//sp_log(LOG_MAIN, LOG_ERR, "cannot create socket: %s", strerror(errno));
	return -1;
    }

    struct ifreq ifr;
    memset(&ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, ifname, IFNAMSIZ-1);
    if(ioctl(fd, SIOCGIFFLAGS, &ifr) == -1) {
	//sp_log(LOG_MAIN, LOG_ERR, "ioctl: %s", strerror(errno));
	close(fd);
	return -1;
    }
    ifr.ifr_flags |= IFF_UP;
    if(ioctl(fd, SIOCSIFFLAGS, &ifr) == -1) {
	//sp_log(LOG_MAIN, LOG_ERR, "ioctl: %s", strerror(errno));
	close(fd);
	return -1;
    }
    close(fd);
    return 0;
}

#define BUFSIZE 8192

static int readNlSock(int sockFd, char *bufPtr, unsigned int seqNum, unsigned int pId)
{
    struct nlmsghdr *nlHdr;
    int msgLen = 0;

    do
    {
	/* Recieve response from the kernel */
	int readLen = recv(sockFd, bufPtr, BUFSIZE - msgLen, 0);
	if(readLen < 0) {
	    return -1;
	}

	nlHdr = (struct nlmsghdr *)bufPtr;

	/* Check if the header is valid */
	if(!NLMSG_OK(nlHdr, (unsigned int)readLen) || (nlHdr->nlmsg_type == NLMSG_ERROR)) {
	    return -1;
	}

	/* Check if the its the last message */
	if(nlHdr->nlmsg_type == NLMSG_DONE) {
	    break;
	} else {
	    /* Else move the pointer to buffer appropriately */
	    bufPtr += readLen;
	    msgLen += readLen;
	}

	/* Check if its a multi part message */
	if((nlHdr->nlmsg_flags & NLM_F_MULTI) == 0) {
	    /* return if its not */
	    break;
	}
    } while((nlHdr->nlmsg_seq != seqNum) || (nlHdr->nlmsg_pid != pId));

    return msgLen;
}

static int parseRoutes(const struct nlmsghdr *nlHdr, uint32_t *gw)
{
    const struct rtmsg *rtMsg;
    const struct rtattr *rtAttr;
    int rtLen;

    rtMsg = (struct rtmsg *)NLMSG_DATA(nlHdr);

    /* If the route is not for AF_INET or does not belong to main routing table then return. */
    if(rtMsg->rtm_family != AF_INET || rtMsg->rtm_table != RT_TABLE_MAIN)
	return 0;

    /* get the rtattr field */
    rtAttr = (struct rtattr *)RTM_RTA(rtMsg);
    rtLen = RTM_PAYLOAD(nlHdr);

    for(;RTA_OK(rtAttr,rtLen);rtAttr = RTA_NEXT(rtAttr,rtLen))
    {
	if(rtAttr->rta_type == RTA_GATEWAY) {
	    memcpy(gw, RTA_DATA(rtAttr), sizeof(*gw));
	    return 1;
	}
    }

    return 0;
}


int get_default_gw(uint32_t *res)
{
    static unsigned int sequence;

    struct {
	struct nlmsghdr  nh;
	struct rtmsg     msg;
	char             attrbuf[512];
    } req;

    memset(&req, 0, sizeof(req));
    req.nh.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
    req.nh.nlmsg_flags = NLM_F_REQUEST | NLM_F_DUMP;
    req.nh.nlmsg_type = RTM_GETROUTE;
    req.nh.nlmsg_seq = ++sequence;
    req.nh.nlmsg_pid = getpid();

    int rtnetlink_socket = socket(AF_NETLINK, SOCK_DGRAM, NETLINK_ROUTE);
    if(rtnetlink_socket == -1) {
	return -1;
    }

    /* Send the request */
    int len = send(rtnetlink_socket, &req, req.nh.nlmsg_len, 0);
    if(len == -1) {
	close(rtnetlink_socket);
	return -1;
    }

    char msgBuf[BUFSIZE];
    struct nlmsghdr *nlMsg = (struct nlmsghdr *)msgBuf;

    /* Read the response */
    if((len = readNlSock(rtnetlink_socket, msgBuf, sequence, getpid())) < 0) {
	close(rtnetlink_socket);
	return -1;
    }
    close(rtnetlink_socket);

    /* Parse and print the response */
    for(;NLMSG_OK(nlMsg, (unsigned int)len);nlMsg = NLMSG_NEXT(nlMsg,len)) {
	// Check if default gateway
	if(parseRoutes(nlMsg, res))
	    return 0;
    }

    return -1;
}

int set_default_gw(uint32_t ip)
{
    int fd = socket(AF_INET, SOCK_DGRAM, 0);
    if(fd == -1)
	return -1;

    struct rtentry route;
    memset(&route, 0, sizeof (struct rtentry));
    struct sockaddr_in *dst, *gw, *mask;
    dst = (struct sockaddr_in *) (&(route.rt_dst));
    gw = (struct sockaddr_in *) (&(route.rt_gateway));
    mask = (struct sockaddr_in *) (&(route.rt_genmask));
    dst->sin_family = AF_INET;
    gw->sin_family = AF_INET;
    mask->sin_family = AF_INET;

    // Set up the data for removing the default route
    dst->sin_addr.s_addr = 0;
    gw->sin_addr.s_addr = 0;
    mask->sin_addr.s_addr = 0;
    route.rt_flags = RTF_UP | RTF_GATEWAY;

    // Remove the default route
    ioctl(fd, SIOCDELRT, &route);

    // Set up the data for adding the default route
    dst->sin_addr.s_addr = 0;
    gw->sin_addr.s_addr = ip;
    mask->sin_addr.s_addr = 0;
    route.rt_metric = 1;
    route.rt_flags = RTF_UP | RTF_GATEWAY;

    // Remove this route if it already exists
    ioctl(fd, SIOCDELRT, &route);

    // Add the default route
    if(ioctl(fd, SIOCADDRT, &route) == -1) {
	close(fd);
	return -1;
    }

    close(fd);
    return 0;
}
