//
// Created by trokay on 18.06.23.
//

#include "protocol.h"
#include "headers.h"


void encode_message(char *data, int size, int *new_size) {
    char *result = (char *) calloc(size * 2, sizeof(char));
    if (result == NULL) {
        printf("Memory allocation failed.\n");
        return;
    }

    int index = 0;
    for (int i = 0; i < size; i++) {
        if (data[i] == 0x7d) {
            result[index++] = 0x7d;
            result[index++] = 0x5d;
        } else {
            if (data[i] == 0x7e) {
                result[index++] = 0x7d;
                result[index++] = 0x5e;
            } else {
                result[index++] = data[i];
            }
        }
    }

    for (int i = 0; i < index; i++) {
        data[i] = result[i];
    }
    *new_size = index;
    free(result);
}

void decode_message(char *data, int size, int *new_size) {
    unsigned char *result = (unsigned char *) calloc(size, sizeof(unsigned char));
    if (result == NULL) {
        printf("Memory allocation failed.\n");
        return;
    }
    int index = 0;
    int flag;
    for (int i = 0; i < size; i++) {
        if (data[i] == 0x7d) {
            flag = 1;
        } else {
            if (data[i] == 0x5e && flag == 1) {
                result[index++] = 0x7e;
                flag = 0;
            } else {
                if (data[i] == 0x5d && flag == 1) {
                    result[index++] = 0x7d;
                    flag = 0;
                } else {
                    result[index++] = data[i];
                }
            }
        }
    }
    for (int i = 0; i < index; ++i) {
        data[i] = result[i];
    }
    (*new_size) = index;
    free(result);
}

char calc_checksum(const char *data, int size) {
    uint8_t tmp = 0;
    while (size > 0) {
        tmp ^= data[size - 1];
        size--;
    }
    return tmp;
}

int extract_data_from_command(char *source, int src_size, char *dest, int *dst_size) {
    char tmp_str[256];
    int tmp = 0, ind = 0;
    for (int i = 0; i < src_size && tmp < 2; i++) {
        if (source[i] == 0x7e) {
            tmp++;
        } else if (tmp == 1) {
            tmp_str[ind++] = source[i];
        }
    }
    if (tmp < 2) {
        return -1;
    }
    *dst_size = ind;

    decode_message(tmp_str, ind, &tmp);

    if (calc_checksum(tmp_str, tmp - 1) != tmp_str[tmp - 1]) {
        return -1;
    }

    for (int i = 1; i < tmp - 1; ++i) {
        dest[i - 1] = tmp_str[i];
    }
    //Command Exec
    //if (tmp_str[0] < 0x80) {

    //}
    tmp -= 2;

    *dst_size = tmp;
    return tmp_str[0];
}


void convert_from_string(char *request,
                         uint8_t *dhcp,
                         uint32_t *ip,
                         uint32_t *netmask,
                         uint32_t *gw,
                         uint32_t *dns) {
    struct config_t config;
    memcpy(&config, request, sizeof(struct config_t));
    memcpy(dhcp, &config.dhcp, sizeof(uint32_t));
    memcpy(ip, &config.ip, sizeof(uint32_t));
    memcpy(netmask, &config.netmask, sizeof(uint32_t));
    memcpy(gw, &config.gw, sizeof(uint32_t));
    memcpy(dns, &config.dns, sizeof(uint32_t));
}

void convert_to_string(char *response, struct config_t *config)
{
  memcpy(response, config, sizeof(struct config_t));
}

/*
 * Получает данные, возвращает строку(ответ) для пересылки UART
 */
void get_response(char response[256],
                  int *response_size,
                  char *data,
                  int data_size,
                  uint8_t resp_code,
                  uint8_t index) {

    int tmp_size = data_size;
    int i = 2;
    response[0] = 0x00;
    response[1] = resp_code;
    //response[2] = index;
    for (; i - 2 < data_size; i++) {
        response[i] = data[i - 2];
    }
    uint8_t cs = calc_checksum(response, i);
    response[i++] = cs;
    encode_message(response, i, &tmp_size);
    response[0] = 0x7e;

    response[tmp_size++] = 0x7e;
    *response_size = tmp_size;
}
